

from warnings_thread_safe import warnings 
from repair_dims import check_dims_simple,check_complex_dims,invalid_string_number

from repair_metadata_prop import check_metadata_entries
from pre_process_module.util_process.util import remove_extra_params,validate_enum,LINE_WITH_TAB,is_number

from util_process.util import LocationType,get_location_label


data_storage = ["uint64Data","doubleData","int64Data","int32Data","floatData","rawData","stringData"]



def check_sparse_complex(tensor,start_error_messege:str="",check_storage_data=False,tensor_id=None):
    
    
    try:
        tmp_name = tensor["values"]["name"]
        tensor_id = f"{tensor_id}-{tmp_name}"
        
    except Exception:
        pass
    
    
    
    
    check=True
    
    
    if "indices" in tensor:

        check = check_tensor_complex(tensor,start_error_messege=f"{start_error_messege} Sparse Initializer (indices) has a wrong format {LINE_WITH_TAB}",check_storage_data=check_storage_data,tensor_id=f"{tensor_id}-indices",omit_name=True)
        
    if check and "values" in tensor:
        if "name" not in tensor["values"]:
            warnings.warn(f"{start_error_messege} \"values\" param needs to have a tensor with \"name\" param")
            return False,None
        check = check_tensor_complex(tensor,start_error_messege=f"{start_error_messege} Sparse Initializer (values) has a wrong format {LINE_WITH_TAB}",check_storage_data=check_storage_data,tensor_id=f"{tensor_id}-values",omit_name=True)
        
    
    if "dims" not in tensor:
        tensor["dims"]=[]
        warnings.warn(f"\n{start_error_messege} \"dims\" param is missing")
    else:
        tensor["shape"]=[]
        
        if not isinstance(tensor["dims"],list):
            return False,None
        for pos,item in enumerate(tensor["dims"]):
            if invalid_string_number(item):
                warnings.warn(f"\n{start_error_messege} \"dim\" entry at pos {pos} is not a integer")
                return False,None
            shape = {"dimValue":int(item),"index":pos,"tensor_id":tensor_id}
            if pos < len(tensor["dims"])-1:
               shape["next_index"]=pos+1
            if pos == len(tensor["dims"])-1:
               shape["last_index"]=""
            tensor["shape"].append(shape)
        del tensor["dims"]
    if "values" not in tensor and "indices" not in tensor:
        warnings.warn(f"\n{start_error_messege} Values and indices are missing")
        return False,None
    
    
    
    
    
    
    remove_extra_params(tensor,["indices","dims","values"],start_error_messege)
    tensor["is_sparse"]=""
    tensor["tensor_id"]=tensor_id
    return True

def check_tensor_complex(tensor,start_error_messege:str="Initializer",check_storage_data=False,tensor_id=None,omit_name=False):
    if not isinstance(tensor,dict):
        
        warnings.warn(f"\n{start_error_messege} Tensor is (it is not a dictionary) -> Tensor its deleted")
        
        return False
    correct=True
    
   
    
    
    
    
    if "dims" in tensor :
         
        if "name" in tensor and not isinstance(tensor["name"],str):
            
            warnings.warn(f"\n{start_error_messege}  Tensor \"name\" param is not a string (sequence of characters)")
            if not omit_name:
                tensor_id = f"{tensor_id}-{tensor["name"]}"
            correct= False
            
        
           
        error = check_dims_simple(tensor["dims"],f"{start_error_messege} Tensor \"dims\" param has a wrong format {LINE_WITH_TAB}",tensor_id=tensor_id)    
        
        if error!=None:
           
            warnings.warn(f"{start_error_messege} error")
            correct= False    
        else:
            dims = tensor["dims"]
            del tensor["dims"]
            tensor["shape"]=dims
            tensor["is_shape"]=""
             
    
        
    if "dataLocation" not in tensor:
        label = get_location_label(LocationType.DEFAULT)
        tensor["dataLocation"] = {"prefLabel":"DEFAULT","altLabel":0,"rdfLabel":label}
        location_type=LocationType.DEFAULT
    else:
        new_entry,location_type = get_correct_data_location(tensor["dataLocation"])
        if new_entry==None:
            warnings.warn(f"\n{start_error_messege} The \"dataLocation\" param is not a valid Location Type")
            correct= False
        else:
           tensor["dataLocation"]=new_entry 
        
    if "dataType" not in tensor:
        
        warnings.warn(f"\n{start_error_messege} Tensor \"dataType\" param is missing")
        
        correct= False
    else:
        enum = check_tensor_data_type(tensor["dataType"])   
        if enum==None:
            warnings.warn(f"\n{start_error_messege} Tensor \"dataType\" param is not a valid DataType")
            correct= False
        else:
            label = get_label_data_type(enum)
            tensor["dataType"]={"prefLabel":enum.name,"altLabel":enum.value,"rdfLabel":label}
            storage_check = validate_data_type_with_storage(enum,tensor,check_storage_data,location_type==LocationType.EXTERNAL,
                                                            f"{start_error_messege} Tensor \"dataType\" param has a incorrect_storage variable {LINE_WITH_TAB}") 
            correct = correct and storage_check
            
      
    if "segment" in tensor:
        check = check_segment(tensor["segment"],start_error_messege=start_error_messege)
        if not check:
            correct=False
        else:
            tensor["segment"]["tensor_id"]=tensor_id
            tensor["is_segment"]=""
        
    
    if "metadataProps" in tensor:
        check,_ = check_metadata_entries(tensor["metadataProps"],start_error_messege=f"{start_error_messege} Tensor \"metadata_prop\" param has a wrong format {LINE_WITH_TAB}",keys=[],element_id=tensor_id)    
        if not check:
            tensor["metadataProps"]=[]
    
    if "externalData" in tensor:
        _,external_correct = check_metadata_entries(tensor["externalData"],start_error_messege=f"{start_error_messege} Tensor \"externalData\" param has a wrong format {LINE_WITH_TAB}",element_id=tensor_id,uri_id="external") 
        correct = correct and external_correct
        
    if "docString" in tensor and not isinstance(tensor["docString"],str):
        
        warnings.warn(f"\n{start_error_messege} The \"docString\" is not a string -> Deleting \"docString\"")

    if correct:
        params = ["externalData","dataLocation","dataType","name","shape","segment","docString","metadataProps","tensor_id","storage_type","value","is_shape","is_segment"]
        params.extend(data_storage)
        
        remove_extra_params(tensor,params,start_error_messege=f"{start_error_messege} The tensor ")
    tensor["is_tensor"]=""   
    tensor["tensor_id"]=tensor_id
    
    
    
     
    return correct



def check_segment(segment,start_error_messege=""):
    correct=True
    if not isinstance(segment,dict):
        warnings.warn(f"{start_error_messege} The \"segment\" param is not a dict")
        return False
    if "begin" not in segment:
        warnings.warn(f"{start_error_messege} The \"segment\" param is missing the \"begin\" param")
        correct=False
    elif invalid_string_number(segment["begin"]):  
        warnings.warn(f"{start_error_messege} The \"segment\" param its \"begin\" param is not a integer")
        correct=False
    elif isinstance(segment["begin"],str):
        segment["begin"]=int(segment["begin"])
        
        
    if "end" not in segment:
        warnings.warn(f"{start_error_messege} The \"segment\" param is missing the \"end\" param")
        correct=False
    elif invalid_string_number(segment["end"]):  
        warnings.warn(f"{start_error_messege} The \"segment\" param its \"end\" param is not a integer")
        correct=False
    elif isinstance(segment["end"],str):
        segment["end"]=int(segment["end"])
    return correct    


def get_correct_data_location(entry):
    label_ext= get_location_label(LocationType.EXTERNAL)
    external = {"prefLabel":"EXTERNAL","altLabel":1,"rdfLabel":label_ext}
    label_def = get_location_label(LocationType.DEFAULT)
    default = {"prefLabel":"DEFAULT","altLabel":0,"rdfLabel":label_def}
    
    
    if isinstance(entry,str) and is_number(entry) or isinstance(entry,int):
        if int(entry)==0:
            return default,LocationType.DEFAULT
        if int(entry)==1:
           return external,LocationType.EXTERNAL 
        return None
    if isinstance(entry,str):
        
        if entry=="DEFAULT":
            return default,LocationType.DEFAULT
        if entry=="EXTERNAL":
            return external,LocationType.EXTERNAL
    
    return None



def check_tensor_simple(tensor,start_error_messege:str="",tensor_id=""):
    
   
    
    if not isinstance(tensor,dict):
        
        warnings.warn(f"\n{start_error_messege} Tensor is not a dict -> Tensor its deleted")
        
        return False
    correct=True
    if "elemType" not in tensor:
        
        warnings.warn(f"\n{start_error_messege} Tensor \"elemType\" param is missing")
        
        correct= False
    else:
        enum = check_tensor_data_type(tensor["elemType"])   
        if enum==None:
            
            warnings.warn(f"\n{start_error_messege} Tensor \"elemType\" param is not a valid DataType")
            
            correct= False
        elif enum==DataType.UNDEFINED:   
            warnings.warn(f"\n{start_error_messege} Tensor \"elemType\" param cannot be \"UNDEFINED\" or \"0\" for a TypeProto entry ")
            correct= False    
        else:
            del tensor["elemType"]
            label = get_label_data_type(enum)
            tensor["dataType"]={"prefLabel":enum.name,"altLabel":enum.value,"rdfLabel":label}
            
    if "shape" in tensor :
        error,dim_present = check_complex_dims(tensor["shape"],f"{start_error_messege} Tensor \"shape\" param has a wrong format {LINE_WITH_TAB}",tensor_id=tensor_id)  
        tensor["is_shape"]=""
        if error!=None:
            warnings.warn("\n"+error)
            correct= False    
        elif dim_present:
            tensor["shape"]=tensor["shape"]["dim"]
        else:
            del tensor["shape"]
    if correct:
        remove_extra_params(tensor,["dataType","shape","is_shape"],start_error_messege=f"{start_error_messege} The tensor ")
    tensor["tensor_id"]=tensor_id    
    return correct
from enum import Enum

class DataType(Enum):
    UNDEFINED = 0
    # Basic types.
    FLOAT = 1        # float
    UINT8 = 2        # uint8_t
    INT8 = 3         # int8_t
    UINT16 = 4       # uint16_t
    INT16 = 5        # int16_t
    INT32 = 6        # int32_t
    INT64 = 7        # int64_t
    STRING = 8       # string
    BOOL = 9         # bool

    # IEEE754 half-precision floating-point format (16 bits wide).
    # This format has 1 sign bit, 5 exponent bits, and 10 mantissa bits.
    FLOAT16 = 10

    DOUBLE = 11
    UINT32 = 12
    UINT64 = 13
    COMPLEX64 = 14   # complex with float32 real and imaginary components
    COMPLEX128 = 15  # complex with float64 real and imaginary components

    # Non-IEEE floating-point format based on IEEE754 single-precision
    # floating-point number truncated to 16 bits.
    # This format has 1 sign bit, 8 exponent bits, and 7 mantissa bits.
    BFLOAT16 = 16

    # Non-IEEE floating-point format based on papers
    # FP8 Formats for Deep Learning, https://arxiv.org/abs/2209.05433,
    # 8-bit Numerical Formats For Deep Neural Networks, https://arxiv.org/pdf/2206.02915.pdf.
    # Operators supported FP8 are Cast, CastLike, QuantizeLinear, DequantizeLinear.
    # The computation usually happens inside a block quantize / dequantize
    # fused by the runtime.
    FLOAT8E4M3FN = 17   # float 8, mostly used for coefficients, supports nan, not inf
    FLOAT8E4M3FNUZ = 18 # float 8, mostly used for coefficients, supports nan, not inf, no negative zero
    FLOAT8E5M2 = 19     # follows IEEE 754, supports nan, inf, mostly used for gradients
    FLOAT8E5M2FNUZ = 20 # follows IEEE 754, supports nan, not inf, mostly used for gradients, no negative zero

    # 4-bit data-types
    UINT4 = 21  # Unsigned integer in range [0, 15]
    INT4 = 22
    FLOAT4E2M1 = 23
    
    
FLOAT_TYPES=[DataType.BFLOAT16,DataType.DOUBLE,DataType.FLOAT,DataType.FLOAT16,DataType.FLOAT8E4M3FN,DataType.FLOAT8E4M3FNUZ,DataType.FLOAT8E5M2FNUZ,DataType.FLOAT8E5M2,DataType.BFLOAT16,DataType.FLOAT4E2M1]
INT_TYPES = [DataType.INT4,DataType.INT8,DataType.INT16,DataType.INT32,DataType.INT64]
UINT_TYPES = [DataType.UINT4,DataType.UINT8,DataType.UINT16,DataType.UINT16,DataType.UINT64]
COMPLEX_TYPES=[DataType.COMPLEX64,DataType.COMPLEX128]


INT_32_TYPES = [DataType.INT32, DataType.INT16, DataType.INT8, DataType.INT4, DataType.UINT16, DataType.UINT8, 
                DataType.UINT4, DataType.BOOL, DataType.FLOAT16, DataType.BFLOAT16, DataType.FLOAT8E4M3FN, 
                DataType.FLOAT8E4M3FNUZ, DataType.FLOAT8E5M2, DataType.FLOAT8E5M2FNUZ,DataType.FLOAT4E2M1]

LABELS_DATA={DataType.FLOAT,DataType.UNDEFINED,DataType.BOOL,DataType.DOUBLE,DataType.STRING}

FLOAT4E2M1_VALUES = {
    0.0, 0.5, 1.0, 1.5, 2.0, 3.0, 4.0, 6.0,
   -0.0, -0.5, -1.0, -1.5, -2.0, -3.0, -4.0, -6.0
}

ONE_ELEMENT = FLOAT_TYPES.copy()
ONE_ELEMENT.extend(INT_TYPES)
ONE_ELEMENT.extend(UINT_TYPES)
ONE_ELEMENT.extend([DataType.STRING,DataType.BOOL])

def check_tensor_data_type(data_type):
    enum = validate_enum(data_type,DataType)
    return enum



def get_label_data_type(data_type:DataType):
    
    if data_type in UINT_TYPES or data_type==DataType.BFLOAT16:
        return data_type.name[0].upper() +data_type.name[1].upper()+ data_type.name[2:].lower()
    if data_type in LABELS_DATA:
        return "Data" +data_type.name[0].upper()+ data_type.name[1:].lower()
        
    return data_type.name[0].upper() + data_type.name[1:].lower()

def get_storage_value(tensor,start_error_messege:str=""):
    tensor_type=None
   
    for storage_type in data_storage:
        if storage_type in tensor  and tensor_type==None:
            tensor_type=storage_type
        elif storage_type in tensor:
            warnings.warn(f"{start_error_messege} There cannot be two different storage parms at the time ({tensor_type}) and ({storage_type})")
            return None,True
    return tensor_type,False    

def change_upper_by_underscore(storage:str):
    result = []
    
    for i, char in enumerate(storage):
        if char.isupper() and i != 0:
            result.append('_')
            result.append(char.lower())
        else:
            result.append(char)
    
    return ''.join(result)


def validate_data_type_with_storage(data_type:DataType,tensor,check_storage_data=False,is_external=False,start_error_messege:str=""):
    
    correct=True
    storage,repeat = get_storage_value(tensor)
    if repeat:
        return False
    default_error = f"\n{start_error_messege} The storage in tensor is {storage} but the type (\"{data_type.name}\") is not valid for that storage"
    missing_error = f"\n{start_error_messege} The storage in tensor is missing"
    
    if "rawData" == storage and (data_type == DataType.UNDEFINED or data_type==DataType.STRING) :
        warnings.warn(default_error)
        return False 
    
    if "floatData" == storage and data_type not in [DataType.FLOAT,DataType.COMPLEX64]:
        warnings.warn(default_error)
        return False
      
    if "int32Data" == storage and data_type not in INT_32_TYPES:
        warnings.warn(default_error)
        return False    
        
    if "stringData"== storage and data_type != DataType.STRING:
        warnings.warn(default_error)
        return False    
    if "int64Data" == storage and data_type != DataType.INT64:
        warnings.warn(default_error)
        return False    
    if "doubleData"== storage and data_type not in [DataType.COMPLEX128,DataType.DOUBLE]:
        warnings.warn(default_error)
        return False    
    if "uint64Data"== storage and data_type not in [DataType.UINT32,DataType.UINT64]:
        warnings.warn(default_error)
        return False
    
    if storage!=None:
        tensor["storage_type"]=change_upper_by_underscore(storage)
           
    
    if check_storage_data:
        
        if storage==None:
            if "name" in tensor and tensor["name"]=="const_tensor" or is_external:
                return True
            warnings.warn(missing_error)
            return False
       
        correct = validate_data_type_with_storage_with_data(data_type,tensor,storage,start_error_messege)
        tensor["value"]=tensor[storage]
        del tensor[storage]
            
    else:
        tensor["value"]=""
        if storage!=None:
            del tensor[storage]
    return correct




def validate_float(data_type:DataType,values:list,default_error,size_error):
    
    
    for value in enumerate(values):
        if isinstance(value,list):
            validate_float(data_type,value,default_error,size_error)
        try:
            for idx,value in enumerate(values):
                type_value = float(value)
        except Exception:
            warnings.warn(default_error)
            return False
        if not check_float_size(type_value,data_type):
            warnings.warn(size_error)
            return False
    return True

def validate_int(data_type:DataType,values:list,default_error,size_error):
    
    
    for value in enumerate(values):
        if isinstance(value,list):
            validate_int(data_type,value,default_error,size_error)
        try:
            for idx,value in enumerate(values):
                int_value = int(value)
        except Exception:
            warnings.warn(default_error)
            return False
        if data_type in INT_TYPES and not check_int_size(int_value,data_type):
            warnings.warn(size_error)
            return False
        if data_type in UINT_TYPES and not check_uint_size(int_value,data_type):
            warnings.warn(size_error)
            return False
    return True

def validate_generic(data_type:DataType,values:list,func,default_error):
    
    
    for value in enumerate(values):
        if isinstance(value,list):
            validate_generic(data_type,value,func,default_error)
        try:
            correct = func(data_type,value)
        except Exception:
            warnings.warn(default_error)
            return False
        if not correct:
            warnings.warn(default_error)
            return False
    return True
            
        


def validate_data_type_with_storage_with_data(data_type:DataType,tensor,storage:str,start_error_messege:str=""):
    #TODO: check types compatiblity (raw_data,shapes correctly,complex_types)
    default_error =  f"\n{start_error_messege} The type in tensor is {data_type.name} but the value at the storage \"{storage}\" is invalid for the type {data_type.name}"
    list_error =  f"\n{start_error_messege} The type in tensor is {data_type.name} but one of the values at the storage \"{storage}\" is invalid for the type {data_type.name}"
    size_error =  f"\n{start_error_messege} The type in tensor is {data_type.name} but the value at the storage \"{storage}\" has a incorrect size for the type {data_type.name}"
    size_error_list =  f"\n{start_error_messege} The type in tensor is {data_type.name} but one of the values at the storage \"{storage}\" has a incorrect size for the type {data_type.name}"
    
   
    if storage=="rawData":
        #TODO: more complex checking
        return True
   

    if not isinstance(tensor[storage],list):
        values = [tensor[storage]]
        is_list=False
        
    else:
        values = tensor[storage]
        
        is_list=True
        
    if data_type in FLOAT_TYPES:
        if is_list:
            return validate_float(data_type,values,list_error,size_error_list)
        else:
            return validate_float(data_type,values,default_error,size_error)
        
  
    if data_type in INT_TYPES or data_type in UINT_TYPES:
        if is_list:
            return validate_int(data_type,values,list_error,size_error_list)
        else:
            return validate_int(data_type,values,default_error,size_error)
        
    if data_type == DataType.STRING:
        def validate_str(_,value):
            return not isinstance(value,str)
        if is_list:
            return validate_generic(data_type,values,validate_str,list_error)
        else:
            return validate_generic(data_type,values,validate_str,default_error)


    
    if data_type == DataType.BOOL:
        
        warnings.warn(default_error)
        def validate_bool(_,value):
            if isinstance(value,bool) or isinstance(value,int) and (value==0 or value==1) :
                return True
            return False
        if is_list:
            return validate_generic(data_type,values,validate_bool,list_error)
        else:
            return validate_generic(data_type,values,validate_bool,default_error)
   
    if data_type in COMPLEX_TYPES:
        #TODO check complex_types
        return True 
    
    
    raise RuntimeError(f"\n{start_error_messege} The type in tensor is {data_type.name} but is not being processed UNEXPECTED ERROR")
    
import numpy as np

def check_int_size(value:int,type:DataType):
    

    if type == DataType.INT4:
        return -8 <= value <= 7
    if type == DataType.INT8:
        return np.iinfo(np.int8).min <= value <= np.iinfo(np.int8).max
    elif type == DataType.INT16:
        return np.iinfo(np.int16).min <= value <= np.iinfo(np.int16).max
    elif type == DataType.INT32:
        return np.iinfo(np.int32).min <= value <= np.iinfo(np.int32).max
    elif type == DataType.INT64:
        return np.iinfo(np.int64).min <= value <= np.iinfo(np.int64).max
    else:
        return False  
    
def check_uint_size(value: int, type: DataType) -> bool:
    if type == DataType.UINT4:
        return 0 <= value <= 15
    elif type == DataType.UINT8:
        return np.iinfo(np.uint8).min <= value <= np.iinfo(np.uint8).max
    elif type == DataType.UINT16:
        return np.iinfo(np.uint16).min <= value <= np.iinfo(np.uint16).max
    elif type == DataType.UINT32:
        return np.iinfo(np.uint32).min <= value <= np.iinfo(np.uint32).max
    elif type == DataType.UINT64:
        return np.iinfo(np.uint64).min <= value <= np.iinfo(np.uint64).max
    else:
        return False
import math
def is_valid_bfloat16(value):
    try:
        f = float(value)
        return math.isfinite(f) and abs(f) <= 3.38953139e+38
    except (ValueError, TypeError):
        return False

def check_float_size(value: float, type: DataType) -> bool:
    if type == DataType.BFLOAT16:
        return is_valid_bfloat16(value)
    elif type == DataType.DOUBLE:
        return np.finfo(np.float64).min <= value <= np.finfo(np.float64).max
    elif type == DataType.FLOAT:
        return np.finfo(np.float32).min <= value <= np.finfo(np.float32).max
    elif type == DataType.FLOAT16:
        return np.finfo(np.float16).min <= value <= np.finfo(np.float16).max
    elif type == DataType.FLOAT8E4M3FN:
        return -128 <= value <= 127 or np.isnan(value)
    elif type == DataType.FLOAT8E4M3FNUZ:
        if value == -0:
            return False  
        return -128 <= value <= 127 or np.isnan(value)
    elif type == DataType.FLOAT8E5M2:
        return np.finfo(np.float32).min <= value <= np.finfo(np.float32).max or np.isnan(value) or np.isinf(value)
    elif type == DataType.FLOAT8E5M2FNUZ:
        if np.isinf(value):
            return False 
        return np.finfo(np.float32).min <= value <= np.finfo(np.float32).max or np.isnan(value)
    elif type == DataType.FLOAT4E2M1:
        return value in FLOAT4E2M1_VALUES

    return False



def is_valid_float4e2m1(value: float) -> bool:
    return value in FLOAT4E2M1_VALUES
    
    
def check_list_data_type(tensor,type:DataType,storage_param:str,is_list:bool,start_error_messege:str=""):    
    if is_list and not isinstance(tensor[storage_param],list):
        warnings.warn(f"\n{start_error_messege} The storage in tensor is {type.name} but in the storage {storage_param} param the value is not a list ")
        return False
    if not is_list and isinstance(tensor[storage_param],list) and len(is_list)>0:
        warnings.warn(f"\n{start_error_messege} The storage in tensor is {type.name} but in the storage {storage_param} param the value is a list with more than one element ")    
        return False
    return True


   
 


from warnings_thread_safe import warnings 

LINE_WITH_TAB="\n\t-> "

DOUBLE_LINE_WITH_TAB="\n\t\t- "

def create_unique_names(already_names,base_name):
    test_name=base_name
    
    idx=1
    while(test_name in already_names):
        test_name=f"{base_name}_{idx}"
        idx=idx+1
    #test_name is new_name, base_name was the first name to try (original name)    
    return test_name

def remove_extra_params(entry:dict,keep_params:list[str],start_error_messege:str=""):
    keys=entry.keys()
    remove_keys=[]
    for key in keys:
        if key not in keep_params:
            remove_keys.append(key)
    accumulate_error = f"{start_error_messege} has extra invalid params {DOUBLE_LINE_WITH_TAB}"
    
    for idx,key in enumerate(remove_keys):
        accumulate_error = f"{accumulate_error} The ({key}) is not a valid parameter. Being removed "
        if idx <len(remove_keys)-1:
            accumulate_error = f"{accumulate_error}{DOUBLE_LINE_WITH_TAB}"
        
        entry.pop(key)
    if remove_keys!=[]:
        warnings.warn(accumulate_error)
        
        
def is_number(s:str):
    try:
        int(s)
        return True
    except ValueError:
        return False
from enum import Enum

def validate_enum(data_type,enum_class:Enum):
    
    if not isinstance(data_type,str) and not isinstance(data_type,int):
        return None

    enum_code = None
    enum_label = None

    if (isinstance(data_type,int) or isinstance(data_type,str) and is_number(data_type)) :
        enum_code=int(data_type)
        
    
    if (isinstance(data_type,str) and enum_code==None) :
        enum_label = data_type
        
        
    if enum_code and int(enum_code) not in enum_class._value2member_map_:
        return None
    elif enum_code:
        return enum_class._value2member_map_[enum_code]

    if enum_label and enum_label not in enum_class._member_names_:
        return None    
    elif enum_label:
        return enum_class[enum_label]
    return None

class LocationType(Enum):
    DEFAULT = 0
    EXTERNAL = 1        
def get_location_label(type:LocationType):
    return "Location"+type.name[0].upper()+type.name[1:].lower()
    
   
class OperatorStatus(Enum):
    EXPERIMENTAL = 0
    STABLE = 1        
    
def get_operator_status_label(type:OperatorStatus):
    return "Status"+type.name[0].upper()+type.name[1:].lower()


def treat_special_chars(input:str):
    treated_result = input.replace("/", "//")
    return treated_result




from warnings_thread_safe import warnings 
import threading

# Custom warning class that includes thread name
class ThreadWarning(UserWarning):
    def __init__(self, message):
        thread_name = threading.current_thread().name
        super().__init__(message)
        self.thread_name = thread_name

# Replacement for warnings.warn
def warn(message, category=None, stacklevel=1, source=None):
    # Wrap the message in ThreadWarning
    if category is None:
        category = ThreadWarning
    # If message is a string, convert to warning object with thread info
    if isinstance(message, str):
        message = ThreadWarning(message)
    warnings.warn(message, category=category, stacklevel=stacklevel+1, source=source)
    



from warnings_thread_safe import warnings 
from pre_process_module.repair_types.repair_metadata_prop import check_metadata_entries
from pre_process_module.repair_types.repair_attributes import check_attributes
from pre_process_module.repair_types.reapair_value_info import check_value_infos

from pre_process_module.util_process.util import LINE_WITH_TAB

# repair params not related to domain_opset (name,domain,version), 
#this includes params the params (inputs,outputs,attribute,attributeProto,value_info,metadata)
def get_func_element_id(function):
    element_id=f"name-{function["name"]}"
    if function["domain"]!="":
        element_id=f"{element_id}-domain-{function["domain"]}"
    if function["overload"]!="":
        element_id=f"{element_id}-overload-{function["overload"]}"
    
    return element_id


def repair_function(function,func_pos):
    
    element_id=f"Func-name-{function["name"]}"
    if function["domain"]!="":
        element_id=f"{element_id}-domain-{function["domain"]}"
    if function["overload"]!="":
        element_id=f"{element_id}-overload-{function["overload"]}"
    
    function["element_id"]=f"{get_func_element_id(function)}"
    
    all_correct=True
    
    all_correct = check_fields_functions(function)
    for idx,input in enumerate(function["input"]):
        if not isinstance(input,str):
            all_correct=False
            warnings.warns(f"The function {function["name"]} defined at pos {func_pos} has a wrong input variable{LINE_WITH_TAB} The input_entry at pos {idx} is not a string {LINE_WITH_TAB} Removing Function")
    for idx,output in enumerate(function["output"]):
        if not isinstance(output,str):
            all_correct=False
            warnings.warns(f"The function {function["name"]} defined at pos {func_pos} has a wrong \"output\" variable {LINE_WITH_TAB} The output_entry at pos {idx} is not a string {LINE_WITH_TAB} Removing Function")
    attribute_list=[]
    if "attribute" in function:
        if not isinstance(function["attribute"],list):
            all_correct=False
            warnings.warns(f"The function {function["name"]} defined at pos {func_pos} has a wrong \"attribute\" variable It is not a list {LINE_WITH_TAB} Removing Function")
        else:
            for idx,attr in enumerate(function["attribute"]):
                if not isinstance(attr,str):
                    all_correct=False
                    warnings.warns(f"The function {function["name"]} defined at pos {func_pos} has a wrong \"attribute\" variable {LINE_WITH_TAB} The attribute_entry at pos {idx} is not a string {LINE_WITH_TAB} Removing Function")
                else:
                    attribute_list.append(attr)
    if "docString" in function and not isinstance(function["docString"],str):
        warnings.warns(f"The function {function["name"]} defined at pos {func_pos} has a wrong format {LINE_WITH_TAB} The \"docString\" param is not a string -> Removing \"docString\"")
        function["docString"]=""
        
    if "metadataProps" in function:
        correct,_ =check_metadata_entries(function["metadataProps"],start_error_messege=f"\nThe function {function["name"]} defined at pos {func_pos} has a wrong metadata_prop param{LINE_WITH_TAB}",keys=[],element_id=element_id)
        
        if not correct:
            function["metadataProps"]=[]
            
    if "attributeProto" in function:
        _,all_correct = check_attributes(function["attributeProto"],start_error_messege=f"\nThe function {function["name"]} defined at pos {func_pos} has a wrong attributeProto variable{LINE_WITH_TAB}",element_id=element_id)
        if all_correct:
            att_names = {item["name"]: item for item in function["attributeProto"]}
            intersect = list(set(att_names) & set(attribute_list))
            if len(intersect)>0:
                all_correct=False
                warnings.warn(f"The function {function["name"]} defined at pos {func_pos} has a wrong format {LINE_WITH_TAB} The \"attributeProto\" param has attributes with the same name with the \"attribute\" param {LINE_WITH_TAB} Removing Function")
    
       
    if "valueInfo" in function:
        correct = check_value_infos(function["valueInfo"],start_error_messege=f"\nThe function {function["name"]} defined at pos {func_pos} has a wrong value_info variable{LINE_WITH_TAB}",element_id=f"{element_id}-valuesInfo")
        if not correct:
            function["valueInfo"]=[]
    else:
        function["valueInfo"]=[]

    connect_input_output(function["valueInfo"],function["input"],function["output"])
    
    return all_correct


def connect_input_output(values:list,inputs:list,outputs:list):
    for value in values:
        if value["name"] in inputs:
            value["is_input"]=value["element_id"]
        elif value["name"] in outputs:
            value["is_output"]=value["element_id"]
        else:
            value["is_value"]=value["element_id"]
        
            

def check_fields_functions(function):
    if "opsetImport" not in function or not isinstance(function["opsetImport"],list):
            function["opsetImport"]=[]
    if "node" not in function or not isinstance(function["node"],list):
            function["node"]=[]
    if "input" not in function or not isinstance(function["input"],list):
            function["input"]=[]
    if "output" not in function or not isinstance(function["output"],list):
            function["output"]=[]

    return True


from pre_process_module.util_process.operators import is_domain_ONNX,get_domain_ONNX_version,is_onnx_operator
from pre_process_module.util_process import ONNX_DOMAIN    
from pre_process_module.util_process.operators import is_opertor_list_func_type,is_register_domain,is_register_operator

from pre_process_module.util_process.data import get_opset_data,function_data_find,opset_data_find,get_functions_data,opset_key,function_key
from warnings_thread_safe import warnings 
from check_functions import repair_function,check_fields_functions
from pre_process_module.util_process.util import remove_extra_params

def repair_opset_and_funcs(data):
    '''The `repair_opset_and_funcs` function in Python is designed to repair and validate opset and
    function data structures by checking for duplicates, missing fields, and conflicting domain
    versions.
    
    Parameters
    ----------
    data
        The `repair_opset_and_funcs` function seems to be a complex function that deals with repairing and
    validating data related to opsets and functions. It performs various checks and repairs on the input
    data to ensure consistency and correctness.
    
    '''
   
    
    
    func_data = __repair_global_opset__(data)
    global_import = get_opset_data(data["opsetImport"])
    add_opset_element_id(data["opsetImport"])
    
    
    for function in data["functions"]:
        __repair_local_opset__(function["opsetImport"],function["name"],func_data,global_import)
        add_opset_element_id(function["opsetImport"])
        
    
def add_opset_element_id(opsets):
    for opset in opsets:
        if opset["domain"]=="":
            opset["element_id"]=f"v{opset["version"]}" 
        else:
            opset["element_id"]=f"domain-{opset["domain"]}-v{opset["version"]}"       
              
        
def __repair_local_opset__(local_import,func_name,func_data,global_import):
    
    domains_opset={}
    
    
    #revisamos opsets locales a funcion buscando duplicados o valores incorrectos
    for idx,opset in enumerate(local_import):
        if __is_non_existance_opset__(opset):
            del local_import[idx]
            continue
        
        __put_onnx_domain__(opset)
        if is_domain_ONNX(opset["domain"]) and "version" not in opset:
            opset["version"]=get_domain_ONNX_version(opset["domain"])
        if not __process_duplicate_opset__(opset,domains_opset):
            del local_import[idx]
            continue
        
        if not __repair_opset__(opset,func_data,func_name=func_name,global_import=global_import):
            del local_import[idx]
            continue
         # no existe ninguna funcion con el dominio -> estamos tratando con un opset que no hace referencia a ningun dominio onnx o funcion
        
        remove_extra_params(opset,keep_params=["domain","version","element_id"],start_error_messege=f"The opset at pos {idx}")   
    __add_base_onnx_opset__(local_import,domains_opset)

def __repair_global_opset__(data):
    
    domains_opset={}
    all_domains=[]
    
    functions_repeat={}
    
    
    # eliminar opsets duplicados con misma version
    # dominio no existente -> dominio ""
    # se guardan los dominios en all_domains
    
    for idx,opset in enumerate(data["opsetImport"]):
        if __is_non_existance_opset__(opset):
            del data["opsetImport"][idx]
            continue
        __put_onnx_domain__(opset)
        
        if is_domain_ONNX(opset["domain"])  and "version" not in opset:
            opset["version"]=get_domain_ONNX_version(opset["domain"])
        elif "version" not in opset:
            opset["version"]="1"
        
        if not __process_duplicate_opset__(opset,domains_opset):
            del data["opsetImport"][idx]
            continue
        if opset["domain"] not in all_domains:
            all_domains.append(opset["domain"])
        remove_extra_params(opset,keep_params=["domain","version","element_id"],start_error_messege=f"The opset at idx {idx}")
        
            
    # onnx domain (base) siempre se incluye si es posible 
    
    __add_base_onnx_opset__(data["opsetImport"],domains_opset)   
            
    
    #revisamos dominios respecto a funciones que tengan las versiones correctas entre si
    # - funciones con nombres de operadores onnx (mas concretamente si son no funcionales (tipo concreto OPERADOR ONNX) ) y el dominio onnx ->invalido
    # - funciones sin nombre -> invalido (añadir un nombre nuevo no seria valido ya que no existirian nodos del grafo que 
    # referencien a esta funcion)
    # - funciones no son un dict ->invalido
    
    # introducimos domian/version de funciones en import opset si no puesto anteriormente (solamente si no son posibles operadores onnx)
    # añadimos dominio nuevo en caso de faltante (solo en el caso que el )

        
            
    #revisamos las funciones
  
    
    
    for idx,function in enumerate(data["functions"]):
        
        if not isinstance(function, dict):
            warnings.warn(f"Function number {idx} has a incorrect format -> function {idx} is being remove",category=SyntaxWarning)
            del data["functions"][idx]
            
            continue
        
        
        if "name" not in function :
            warnings.warn(f"Function number {idx} has no name it cannot be identified -> function {idx} is being remove",category=SyntaxWarning)
            del data["functions"][idx]
            
            continue
        
        
        
        name = function["name"]
        
        domain=None
        
        
        
        if "domain" in function:
            domain = function["domain"]
        
        opsets_filter_domain = opset_data_find(domains_opset,domain)
        
        # Eliminamos
        func_operators=[]
        if "domain" in function and is_register_domain(function["domain"]):
            check=False
            
            check,func_operators = is_function_contradictory(name,domain,domains_opset)
            
            if check:
                warnings.warn(f"\nFunction number {idx} with name {name} and domain {domain} Enters with conflict with predefined operators. \n"+
                              f"(which are not Function type and cannot have function implementation) -> Function number {idx} is being remove")
                del data["functions"][idx]
                
                continue
        
        if "domain" not in function and is_register_operator(name):
            check=False
            
            check,func_operators = is_function_contradictory(name,domains_opset=domains_opset)
            
            if check:
                warnings.warn(f"\nFunction number {idx} with name {name} and no domain Enters with conflict with predefined operators. \n"+ 
                              f"(which are not Function type and cannot have function implementation) ->Function number {idx} is being remove")
                del data["functions"][idx]
                
                continue
                
        # añadimos overload vacio en caso de que no exista
        if 'overload' not in function:
            function['overload'] = ""
            overload = ""
        else:
            overload = function['overload']
        
        if "domain" in function and function_key(domain,name,overload) in functions_repeat.keys():    
            warnings.warn(f"\nGRAVE WARNING: Function number {idx} with name {function["name"]} has same domain/name/overload than other defined function ->the second instance is remove")
            
            del data["functions"][idx]
            continue
        
        
        
        if len(func_operators)>0:
            #TODO: caso raro estamos antes funcion que tiene el mismo nombre que un operador predefinido y puede que tenga una 
            # reimplementacion del metodo pero usando una funcion
            __revised_funcion_onnx_incompatible__(data,function)
            # por ahora simplemente borramos la funcion
            warnings.warn("TODO: operadores funcionales y caso de implementacion con su nombre")
            del data["functions"][idx]
            
            continue
           
            
         
        # Añade el dominio y version en opset_imports ya que faltaba (para funciones)
        
        if "domain" in function and len(func_operators)==0 and len(opsets_filter_domain)==0:
            #Add missing domains at opset import for only custom functions   
            new_entry={"domain":function["domain"],"version":"1"}
            data["opsetImport"].append(new_entry)
            domains_opset[opset_key(domain,1)]=new_entry
            

    
        if "domain" in function and len(func_operators)==0:
            # las funciones no deberian tener la version como parametro no obstante lo vamos a añadir/reescribir para que el mapping sea correcto
            # usaremos los valores del opset existente y en caso que no haya version pondremos la version a 1
            new_version=1
            
            
            opsets:dict = opset_data_find(domains_opset,domain=function["domain"])
            
            
            if len(opsets)==0:
                # domain no in domains_opsets this case should not be possible
                warnings.warn("Unexpected case, when obtaining version for functions")
                del data["functions"][idx]
                
                continue
             
            if len(opsets)==1:
                new_version = opsets[0]["version"]
                
            if len(opsets)>1:
                max_version=None
                idx_opset=None
                for idx,opset in enumerate(opsets):
                    if not max_version:
                        max_version=opset["version"]
                        idx_opset=idx
                    else:
                        max_version=max(opset["version"],max_version)
                        #TODO: verificar del opsets correctly
                        del opsets[idx_opset]
                        warnings.warn("Duplicate opsets with difference versions for user functions are not allowed, as we "+
                                      "we cannont identified specific domain the function relates to we have chose the hightest version delete the others ")
                        idx_opset=idx
            
        
            function['version']=new_version
            
        if "domain" in function and function["domain"] not in all_domains:
            #guardamos el dominio de la funcion en caso que no existiese en all_domains
            #TODO: revisar que no entra en conflicto con el caso raro de func_operators
            all_domains.append(function["domain"])
            
        if  "domain" not in function and len(func_operators)==0:
            #no es un operador pero le falta dominio a la funcion
            # añadimos un nuevo dominio que no exista en la lista de all_domains y añadimos su version a 1
            __new_domain_func__(all_domains,data,function)
            
        if "domain" in function and "overload" in function:    
            functions_repeat[function_key(function["domain"],function["name"],function["overload"])]=function
        
        
        
        if not repair_function(function,idx):
            del data["functions"][idx]
            
            continue
        
        
        opset_id =f"v{function['version']}"
        if function['domain']!="":
            opset_id=f"domain-{function['domain']}-{opset_id}"
        function["opset_id"]=opset_id
            
        remove_extra_params(function,keep_params=["name","domain","overload","version","opsetImport","input","output",
                                                  "attribute","attributeProto","metadataProps","docString","opset_id",
                                                  "valueInfo","node","element_id"],start_error_messege=f"The Function ({function["name"]})")  
    
    func_data = get_functions_data(data)
  
    # reparar el resto de opsets 
    
    return func_data




def __is_non_existance_opset__(opset):
    if "domain" not in opset and "version" not in opset:
        warnings.warn("Opset with no domain or version is being delete")
        return True
    return False

def __add_local_results_operators__(operators_local,operators):
    for operator in operators_local:
            operators.add(operator)

def __revised_funcion_onnx_incompatible__(data,function):
    #TODO: revised if function is valid with same name/domain than ONNX operators
    pass

def is_function_contradictory(name,domain=None,domains_opset=None,ignore_func_operators=False):
    '''The function `is_function_contradictory` checks for contradictions in function names and domains.
    
    Parameters
    ----------
    name
        The `name` parameter represents the name of the function for which you want to check for
    contradictions.
    domain
        The `domain` parameter typically refers to the set of input values for a function. It specifies the
    valid inputs that the function can accept.
    domains_opset
        The `domains_opset` parameter likely refers to a set of domains that the function operates on or is
    defined for. This parameter is used to check for contradictions in the function's domain. If you
    provide the specific domains in the `domains_opset` parameter, the function `is_function_contrad
    
    Returns
    -------
        The function `is_function_contradictory` is returning the result of either
    `__is_function_domain_contradictory__` or `__is_function_name_contradictory__` based on the
    conditions provided.
    
    '''
    if domain!=None and domains_opset:
        return __is_function_domain_contradictory__(name,domain,domains_opset,ignore_func_operators=ignore_func_operators)      
    else:
        return __is_function_name_contradictory__(name,ignore_func_operators=ignore_func_operators)

def __is_function_domain__multiple_version_contradictory__(name,domain,opsets:dict):
    operators=set()
    for opset in opsets:
        version = opset["version"]
        _,operators_local = is_onnx_operator(name,domain,version)
        __add_local_results_operators__(operators_local,operators)
    return len(operators)>0,operators  
 
 
 
def __is_function_domain_contradictory__(name,domain,domains_opset,ignore_func_operators=False):
    
    opsets = opset_data_find(domains_opset,domain)
    check=False
    func_operators=[]
    
    if len(opsets)==1:
        version = opsets[0]["version"]
        check,operators = is_onnx_operator(name,domain,version)
        
    if len(opsets)>1:
        check,operators = __is_function_domain__multiple_version_contradictory__(name,domain,opsets)   
    
    if len(opsets)==0:
        # compruebo respecto a todos los operadores registrados independientemente de la version
        check,operators = is_onnx_operator(name,domain)    
        
    # caso especial que sea un operador 
    
    if not ignore_func_operators:
        return check,operators
    
    if check:
        is_func,func_operators = is_opertor_list_func_type(operators)
        check = not is_func # si el
        
    return check,func_operators

def __is_function_name_contradictory__(name,ignore_func_operators=False):
    check,operators = is_onnx_operator(name)
    if not ignore_func_operators:
        return check,operators
    
    if check:
        is_func,func_operators = is_opertor_list_func_type(operators)
        check = not is_func
    return check,func_operators
    

            
                         
def __repair_opset__(opset,func_data,func_name=None,global_import=None):
    

    if "domain" in opset and "version" not in opset and not is_domain_ONNX(opset["domain"]) and len(function_data_find(func_data,domain=opset["domain"]))==0:
        
        if func_name:
            warnings.warn(f"\nDominio {opset["domain"]}, definido en el opset de la funcion {func_name} le faltaba el atributo version y el dominio no es conflictivo -> la version es 1 ahora")
        else:
            warnings.warn(f"\nDominio {opset["domain"]}, definido en el grafo le faltaba el atributo version y el dominio no es conflictivo -> la version es 1 ahora")
        opset["version"]=1
    
    if "domain" in opset and "version" not in opset and len(function_data_find(func_data,domain=opset["domain"]))>0:
        
            
        # import_meta_global contiene informacion de opsets en el nivel global
        
        possible_list = opset_data_find(global_import,domain=opset["domain"])
        # solo hay una combinacion dominio/version
        if len(possible_list)==0 and func_name:
            warnings.warn(f"\nDominio {opset["domain"]}, definido en el opset de la funcion {func_name} no tenia version. El dominio hace referencia a una funcion que existe en la lista de funciones"+
                                " no obstante el dominio no aparece en la lista de opsets global (no es un dominio de una funcion existente).\n Borrando opset entry conflictiva ")
            return False
            
    
        if len(possible_list)==1 and func_name:
            opset["version"]=possible_list[0]["version"]
        
        
        if len(possible_list)>1 and func_name:
            warnings.warn(f"\nDominio {opset["domain"]}, definido en el opset de la funcion {func_name} no tenia version.El dominio hace referencia a una funcion que existe en la lista de funciones"+
                                " no obstante el dominio aparece mas de una vez con versiones distintas (distintas versiones para la misma funcion).\n Borrando opset entry conflictiva ")
            return False
    return True


# crear dominio que no exista en all_domains y añadirlo en el opset_import            
def __new_domain_func__(all_domains,data,function):
    
    idx=1
    new_domain = f"undefined_domain_{idx}"
    
    while(new_domain in all_domains):
        idx=idx+1
        new_domain = f"undefined_domain_{idx}"
    
   
    function["version"]=1     
    new_entry={"domain":new_domain,"version":1}
    data["opsetImport"].append(new_entry)
    function["domain"]=new_domain
    
    all_domains.append(new_domain)          
            
                    

def __put_onnx_domain__(opset):
    if "version" in opset and "domain" not in opset:
        opset["domain"]=ONNX_DOMAIN
        

def __process_duplicate_opset__(opset,domains_opset):
    

    #intentamos eliminar duplicados en los opsets
    
    
    domain=opset["domain"]
    version=opset["version"]
      
    if domain!=None not in domains_opset.keys():
        
        domains_opset[opset_key(domain,version)]=opset
    else:
        warnings.warn("Opset repetidos con mismo domain/version borrando opset repetido")
        return False
    return True
        
                 

def __add_base_onnx_opset__(opset_field,domains_opset):
    
            
    # onnx domain (base) siempre se incluye si es posible
    
    opsets_list = opset_data_find(domains_opset,"")
    if len(opsets_list)==0:
        new_entry= {"domain":"","version":get_domain_ONNX_version(ONNX_DOMAIN)}
        opset_field.append(new_entry) 
        domains_opset[ONNX_DOMAIN]=new_entry 
     
  
   # repairs and checks for all opsetImports on the onnx json file 



from warnings_thread_safe import warnings 
from pre_process_module.util_process.util import remove_extra_params,LINE_WITH_TAB

def check_metadata_entry(entry,keys:list[str],start_error_messege:str=""):
    
    params= ["key","value"]
    
    if not isinstance(entry,dict):
        warnings.warn( f"{start_error_messege} is not a dict")
        return False
    
    if not isinstance(entry,dict):
        warnings.warn( f"{start_error_messege} is not a dict")
        return False
    
    if "key" not in entry and "value" not in entry:
        warnings.warn( f"{start_error_messege} entry has not key or value -> Deleting entry")
        return False
        
    
    if "key" in entry and entry["key"] in keys:
        warnings.warn( f"{start_error_messege} key (\"{entry["key"]}\") is already in use, keys should be distinct -> Deleting Duplicate Key Entry")
        return False
    
    if "key" in entry and "value" not in entry:
        entry["value"]=""
        
    if "key" not in entry and "value" in entry and "" not in keys:
        entry["key"]=""
    elif "key" not in entry and "value" in entry:
        warnings.warn( f"{start_error_messege} its key is missing and the value is filled, but the key (\"\") is already in use")
        return False
    
    if "key" in entry and "value" in entry and not isinstance(entry["key"],str):
        warnings.warn( f"{start_error_messege} its key is not a string")
        return False
        
    if "key" in entry and "value" in entry and not isinstance(entry["value"],str):
        warnings.warn( f"{start_error_messege} its value is not a string")
        return False
    
    
    remove_extra_params(entry,params,start_error_messege)
    return True


def check_metadata_entries(entries,keys:list[str]=None,start_error_messege:str="",element_id="",uri_id="metadata"):
    if start_error_messege==None:
        start_error_messege:str=""
    if keys==None:
        keys=[]
    all_correct=True
    if not isinstance(entries,list):
        warnings.warn(f"{start_error_messege} metadata entries are not a list")
        return False,False
    for idx,entry in enumerate(entries):
        check =check_metadata_entry(entry,keys=keys,start_error_messege=f"{start_error_messege} the entry at pos {idx} has a wrong format {LINE_WITH_TAB}")
        if not check:
            entries.remove(entry)
            all_correct=False
        else:
            keys.append(entry["key"])
            if element_id!="":
                entries[idx]["element_id"]=f"{element_id}-{uri_id}-{entry["key"]}"
            else:
                entries[idx]["element_id"]=f"{uri_id}-{entry["key"]}"
    return len(entries)>0,all_correct
    

       


# threaded_warnings.py

import warnings  as std_warnings
import threading



class WarningsProxy:
    
    def warn(self,message, category=None, stacklevel=1, source=None):
        thread_name = threading.current_thread().name
        if isinstance(message, str):
            message = f"[{thread_name}] {message}"
        if category is None:
            category = UserWarning
        std_warnings.warn(message, category=category, stacklevel=stacklevel + 1, source=source)
    

    def __getattr__(self, attr):
        # Proxy all other attributes to the real warnings module
        return getattr(std_warnings, attr)

# Export this as "warnings"
warnings = WarningsProxy()



from repair_tensor import check_tensor_complex,check_sparse_complex
from warnings_thread_safe import warnings 
from pre_process_module.util_process.util import LINE_WITH_TAB





def repair_initializers(graph,start_error_messege="\n"):
    names=[]
    
    for idx,tensor in enumerate(graph["initializer"]):
        
        
        if "name" in tensor and tensor["name"] not in names:
            name=tensor["name"]
            names.append(name)
            check = check_tensor_complex(tensor,start_error_messege=f"{start_error_messege}Initializer ({name}) at pos {idx} has a wrong format {LINE_WITH_TAB}",tensor_id=f"Graph-{graph["name"]}-Initializer")
        elif "name" in tensor:
            warnings.warn(f"{start_error_messege}Initializer ({name}) at pos {idx}  Has not unique name -> Initializer Being deleted")
            del graph["initializer"][idx]
            continue
        else:
            warnings.warn(f"{start_error_messege}Initializer at pos {idx}  Has no name -> Initializer Being deleted")
            del graph["initializer"][idx]
            continue
        
        if not check:
            if name:
                warnings.warn(f"{start_error_messege}Initializer ({name}) at pos {idx} -> Initializer  Being deleted")
            else:
                warnings.warn(f"{start_error_messege}Initializer at pos {idx} -> Initializer Being deleted")
            del graph["initializer"][idx]
    
    names=[]
    for idx,tensor in enumerate(graph["sparseInitializer"]):
        
        tmp_name=""
        try:
            tmp_name = tensor["values"]["name"]
        except Exception:
            pass
            
        check = check_sparse_complex(tensor,start_error_messege=f"Sparse Initializer ({tmp_name})  at pos {idx} has a wrong format {LINE_WITH_TAB}",tensor_id=f"Graph-{graph["name"]}-SparseInitializer")
        if not check:
            if tmp_name:
                warnings.warn(f"{start_error_messege}Sparse Initializer ({tmp_name}) at pos {idx} -> InitializerBeing deleted")
            else:
                warnings.warn(f"{start_error_messege}Sparse Initializer at pos {idx} -> InitializerBeing deleted")
            del graph["sparseInitializer"][idx]

        if check and tmp_name in names:
            warnings.warn(f"{start_error_messege}Sparse Initializer ({tmp_name}) at pos {idx}  Has not unique name -> Sparse Initializer Being deleted")
            del graph["sparseInitializer"][idx]
        elif check:
            names.append(tmp_name)     
            
        
        
    







from pre_process_module.util_process.util import remove_extra_params,is_number   
    
def invalid_string_number(value):
    return isinstance(value,str) and not is_number(value) or (not isinstance(value,str) and not isinstance(value,int))

def check_dims_simple(dims,start_error_messege:str="",tensor_id=None):
    
    

    if not isinstance(dims,list):
        error_messege=f"{start_error_messege} The \"dims\" param is not a list"
        
        return error_messege
    
    len_dims = len(dims)
    for idx,dim in enumerate(dims):
        if invalid_string_number(dim):
            error_messege=f"{start_error_messege} The dim at pos {idx} of \"dims\" param is not a number"
            return error_messege
        dims[idx]= {"dimValue":int(dim),"index":idx,"last_index":len_dims-1,"next_index":idx+1}
        if tensor_id!=None and isinstance(tensor_id,str):
            dims[idx]["tensor_id"]=tensor_id

    return None


def check_complex_dim(dim,start_error_messege):
   

    if not isinstance(dim,dict):
        error_messege=f"{start_error_messege} not a dict"
        return error_messege
    
    if "dimParam" in dim and "dimValue" in dim:
        error_messege=f"{start_error_messege} has both dimParam and dimValue, only one"
        return error_messege

    if "dimParam" in dim and not isinstance(dim["dimParam"],str):
        error_messege=f"{start_error_messege} dimParam only accepts string values"
        return error_messege

    if "dimValue" in dim and invalid_string_number(dim["dimValue"]):
        error_messege=f"{start_error_messege} dimValue is not a valid number"
        return error_messege
    if "dimValue" in dim and isinstance(dim["dimValue"],str):
        dim["dimValue"]=int(dim["dimValue"])

    
    remove_extra_params(dim,keep_params=["dimParam","dimValue"],start_error_messege=start_error_messege)
    return None


def check_complex_dims(shape,start_error_messege:str="",tensor_id=None):
    
    if start_error_messege==None:
        start_error_messege:str=""
    
    if not isinstance(shape,dict):
        error_messege=f"{start_error_messege} The \"shape\" param is not a dict"
        return error_messege,False
    if "dim" not in shape:
        error_messege=f"{start_error_messege} The \"dim\" param is not in the \"shape\" param"
        #shape can be empty
        return None,False
    if not isinstance(shape["dim"],list):
        error_messege=f"{start_error_messege} The \"dim\" param in \"shape\" param is not a list"
        return error_messege,False
    
    remove_extra_params(shape,keep_params=["dim"],start_error_messege=start_error_messege)
    
    len_dims = len(shape["dim"])
    for idx,dim in enumerate(shape["dim"]):
        
        error_messege = check_complex_dim(dim,f"{start_error_messege} The dim at pos {idx} of \"dims\" param in \"shape\" param")
        if error_messege!=None:
            return error_messege,False

        shape["dim"][idx]["index"]= idx
        shape["dim"][idx]["last_index"]= len_dims-1
        shape["dim"][idx]["next_index"]= idx+1
        if tensor_id!=None and isinstance(tensor_id,str):
            shape["dim"][idx]["tensor_id"]=tensor_id

    return None,True
    






from warnings_thread_safe import warnings 
from pre_process_module.util_process.util import remove_extra_params,LINE_WITH_TAB

from repair_metadata_prop import check_metadata_entries

def check_annotations(graph,element_id,start_error_messege=""):
    if "quantizationAnnotation" not in graph:
        return True
    
    annotation = graph["quantizationAnnotation"]
    tensor_name="emoty_name"
    
    if "tensorName" in annotation and not isinstance(annotation["tensorName"],str):
        warnings.warn(f" {start_error_messege} The param \"tensorName\" is not a string -> Deleting optional param ")

    elif "tensorName" in annotation:
        tensor_name=annotation["tensorName"]
        graph["quantizationAnnotation"]["connects"]=""
    
    if not isinstance(annotation,dict):
        warnings.warn(f" {start_error_messege} The quantizationAnnotation is not a dict -> quantizationAnnotation being deleted ")
        graph["quantizationAnnotation"]=[]
        return False
    
    if "quantParameterTensorNames" in annotation:
        correct,_ = check_metadata_entries(annotation["quantParameterTensorNames"],keys=[],element_id=element_id,uri_id="quant_parameter",
                               start_error_messege=f"{start_error_messege} The param \"quantParameterTensorNames\" has a wrong format on its entries")
        if not correct:
            warnings.warn(f"{start_error_messege} The param \"quantParameterTensorNames\" has no entry remaining -> quantizationAnnotation being deleted")
            graph["quantizationAnnotation"]=[]
        else:
            graph["quantizationAnnotation"]["element_id"]=f"{element_id}-tensor_annotation-{tensor_name}"
            remove_extra_params(graph["quantizationAnnotation"],keep_params=["tensorName","quantParameterTensorNames","element_id"],start_error_messege=f"{start_error_messege} The quantizationAnnotation ")
        return correct
    else:
        warnings.warn(f"{start_error_messege} The param \"quantParameterTensorNames\" is missing -> quantizationAnnotation being deleted")
        graph["quantizationAnnotation"]=[]


from . import ONNX_DOMAIN

def get_opset_data(import_field:dict):
    
    versions={}
    
    
    for item in import_field:
        if "version" in item and "domain" in item:
            versions[opset_key(item["domain"],item["version"])]={"domain":item["domain"],"version":item["version"]}
            
        if "version" in item and "domain" not in item:
            versions[opset_key(ONNX_DOMAIN,item["version"])]={"domain":item["domain"],"version":item["version"]}
            import_field["domain"]=ONNX_DOMAIN 
              
    return versions








def get_functions_data(data):
    functions_data=dict()

    for function in data["functions"]:
        item = function
        functions_data[function_key(function["domain"],function["name"],function["overload"])] = item

    return functions_data

def function_data_find(func_data: dict, domain=None, name=None, overload=None):
    
    return __dynamic_dict_find__(func_data,3,[domain,name,overload])

def opset_data_find(import_data: dict, domain=None, version=None):
    
    
    return __dynamic_dict_find__(import_data,2,[domain,version])

def __dynamic_dict_find__(dict:dict,number_keys,key_values:list[str]):
    
    result = []
    
    
    # Iterar sobre las claves del diccionario
    for key in dict.keys():
        # Descomponer la clave en sus partes (domain, version)
        key_parts = [part.strip('"') for part in key.split('".')]
        
        # Verificar si la clave tiene el formato esperado
        if len(key_parts) != number_keys:
            continue
        
        ignore=False
       
        for idx,key_local in enumerate(key_parts):
            
            if (key_values[idx]!=None) and str(key_values[idx])!=str(key_local):
                ignore=True
        # Comprobar si la clave coincide con los parámetros dados
        
        if ignore:
            continue
        # Si la clave coincide con todos los criterios, agregar el valor al resultado
        result.append(dict[key])
    
    return result


def opset_key(domain,version):
    return __build_dynamic_key__([domain,version])
    
def function_key(domain,name,overload):
    return __build_dynamic_key__([domain,name,overload])


def get_opset_values(opset_data,key):
    return __get_values_dynamic__(opset_data,["domain","name","version"],key)

def get_func_values(func_data,key):
    return __get_values_dynamic__(func_data,["domain","name","overload"],key)


def __get_values_dynamic__(dict,key_names:list,key_name):
    if key_name not in key_names:
        return None
    idx = key_names.index(key_name)
    values=[]
    for key in dict.keys():
        parts=key.split(".")
        if parts[idx] not in values:
            values.append(parts[idx])
    return values

def __build_dynamic_key__(key_values):
    if len(key_values)==0:
        raise RuntimeError("Dynamic key construction receive 0 as number of params")
    key=""
    for idx,key_part in enumerate(key_values):
        if idx < len(key_values)-1:
            key=key+"\""+str(key_part)+"\""+"."
        else:
            key=key+"\""+str(key_part)+"\""
    return key  




from onnx import load_model
import os
from google.protobuf.json_format import MessageToJson
import json
from json_util import save_json,load_json,get_log_paths


def clear_weight_params(tensor):
    
    tensor.ClearField("float_data")
    tensor.ClearField("int32_data")
    if tensor.string_data !=[]:
        tensor.raw_data = ""
    tensor.ClearField("int64_data")
    if tensor.raw_data !=[]:
        tensor.raw_data = b""
    tensor.ClearField("double_data")
    tensor.ClearField("uint64_data")

    

def load_graph_onnx(model_path):
    model = load_model(model_path,load_external_data=False)
    
    for tensor in model.graph.initializer:
        clear_weight_params(tensor)
    for sparse_init in model.graph.sparse_initializer:
        clear_weight_params(sparse_init.indices)
        clear_weight_params(sparse_init.values)
    
    return model


def convert_json(onnx_model):
    s = MessageToJson(onnx_model)
    onnx_json = json.loads(s)
    return onnx_json



    
from datetime import datetime
def process_model(model_path,tmp_path,relative_dir="",log_dirs="logs",cache=False):
    file_name="loaded_model.json"
    if not isinstance(log_dirs,list):
        log_dirs=[log_dirs]
    log_paths=get_log_paths(log_dirs,file_name,relative_dir)
  
            
    is_in_cache = os.path.exists(tmp_path)
            
    if not cache or (not is_in_cache):
        if cache and not os.path.exists(model_path):
            raise ValueError("Cache was activated, but neither cache file or model file were present")
       
        onnx_model = load_graph_onnx(model_path)
        onnx_json = convert_json(onnx_model) 
        
        for log_path in log_paths:
            save_json(onnx_json,log_path)
        
        if not (cache and is_in_cache):
            save_json(onnx_json,tmp_path)
        
                
               
    executed_cache=False        
    if cache and is_in_cache:
        onnx_json = load_json(tmp_path)
        executed_cache=True
        
                  
    return onnx_json,executed_cache